(*
  Copyright (c) 2006, Eric Breck
  All rights reserved.

  Redistribution and use in source and binary forms, with or without
  modification, are permitted provided that the following conditions are met:

  1 Redistributions of source code must retain the above copyright notice, this
    list of conditions and the following disclaimer.

  2 Redistributions in binary form must reproduce the above copyright notice,
    this list of conditions and the following disclaimer in the documentation
    and/or other materials provided with the distribution.

  3 The name of the author may not be used to endorse or promote products
    derived from this software without specific prior written permission.

  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
  DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE
  FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
  DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
  SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
  CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
  OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
  OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *)

(** generate a function to print out an options record *)
let gen_to_string loc fields =
  let prints = List.map (fun (loc,name,mutable_,type_,pvl,default) -> 
    let to_string = try 
      <:expr< $List.assoc "str" pvl$ field>>
    with Not_found ->
      match type_ with
      | `Symbol rfl -> 
          let str_of_sym = List.map (fun (loc,n,s) -> 
            <:patt< ` $uid:n$ >>, None, <:expr< $str:s$ >>
          ) rfl in
          <:expr< match field with [ $list:str_of_sym$ ] >>
      | `Type <:ctyp< int    >> -> <:expr<string_of_int field>>
      | `Type <:ctyp< float  >> -> <:expr<string_of_float field>>
      | `Type <:ctyp< string >> -> <:expr<field>>
      | `Type <:ctyp< unit   >> -> <:expr<"()">>
      | `Type <:ctyp< bool   >> -> <:expr<string_of_bool field>>
      | _ -> <:expr<"(user-specified data)">> in
    let to_string = match default with
      Some _ -> 
      <:expr<let field = record.$lid:name$ in $to_string$>>
    | None ->
      <:expr<match record.$lid:name$ with [ None -> "None"
      | Some field -> $to_string$ ] >> in
    name,to_string
  ) fields in
  let format_string = List.fold_left (fun acc (name,_) ->
    name ^ " = %s\n" ^ acc
  ) "" prints in
  let printf_expr = List.fold_right (fun (_,to_string) acc ->
    <:expr< $acc$ $to_string$ >>
  ) prints <:expr< Printf.sprintf $str:format_string$ >> in
  <:expr<fun record -> $printf_expr$ >>

(* for help string... *)
type actiontype =
  Set | Clear | Store of string | Assign of string | Action | Conversion |
  Handler

(** use Pcaml.string_of and Pr_o to generate a string for expressions.
  * As a special case, for `Symbol types, look up the user-string for them *)
let string_of_expr type_ e =
  match type_ with `Type _ -> Pcaml.string_of Pcaml.pr_expr e
  | `Symbol rfl ->
    match e with <:expr< ` $uid:u$ >> ->
      (try
        let _,_,s = List.find (fun (_,uid,_) -> uid=u) rfl in s
      with Not_found ->
        let values = String.concat "|" (List.map (fun(_,_,s) -> s) rfl) in
        failwith (Printf.sprintf "invalid value %s for type %s" 
          (Pcaml.string_of Pcaml.pr_expr e) values))
    | e -> 
        let values = String.concat "|" (List.map (fun(_,_,s) -> s) rfl) in
        failwith (Printf.sprintf "invalid value %s for type %s" 
          (Pcaml.string_of Pcaml.pr_expr e) values)

(** generate the 'specs' field which contains keys, action and help for every
    set of options *)
let gen_spec (loc,name,mutable_,type_,pvl,default) =
  let namestr = match name with Some name -> name | None -> "UNK" in
  let key = try List.assoc "key" pvl with Not_found -> <:expr<[]>> in
  (* figure out the action function *)
  (* first, see if the user specified a 'handler' function *)
  let actiontype,action = try 
    Handler, <:expr< Handler $List.assoc "handler" pvl$ >> 
  (* next, see if the user specified a 'conversion' function *)
  with Not_found -> try
    let c = List.assoc "conversion" pvl in
    let body = 
      match name with 
      | Some name -> <:expr< {(ap.options) with $lid:name$ = $c$ ap s} >>
      | None -> <:expr< do {$c$ ap s; ap.options} >> in
    Conversion,<:expr< Parseopt.Conversion (fun ap s -> $body$ ) >>
  (* next, see if the user specified an 'action' function *)
  with Not_found -> try 
    let a = List.assoc "action" pvl in 
    let body = 
      match name with 
      | Some name -> <:expr< {(ap.options) with $lid:name$ = $a$ ap} >>
      | None -> <:expr< do {$a$ ap; ap.options} >> in
    let actiontype = match a with 
    | <:expr< set >> -> Set
    | <:expr< clear >> -> Clear
    | MLast.ExApp(_,MLast.ExLid(_,"store"),e) -> 
        Store (string_of_expr type_ e) 
    | _ -> Action in
    actiontype,<:expr< Parseopt.Action (fun ap -> $body$ ) >>
  (* finally, use a function which assigns the specified value to the field *)
  with Not_found ->
    match name with 
      None -> invalid_arg "gen_spec: anonymous fields must specify action "
    | Some name ->
      (let actiontype,action = match type_ with
       | `Symbol rfl -> 
           let values = String.concat "|" (List.map (fun(_,_,s) -> s) rfl) in
           let sym_of_str = List.map (fun (_,n,s) -> 
             <:patt< $str:s$ >>, None, <:expr< ` $uid:n$ >>
           ) rfl in
           let sym_of_str = List.rev ((<:patt< _ >>,None,
           <:expr< raise (Parseopt.ConversionFailure s $str:values$ $str:name$ )>>)::sym_of_str) in
           Assign values, <:expr<match s with [$list:sym_of_str$]>>
       | `Type <:ctyp< int   >> -> Assign "integer",<:expr< try int_of_string s with [ Failure
       "int_of_string" -> raise (Parseopt.ConversionFailure (s,"int",$str:name$)) ] >>
       | `Type <:ctyp< float >> -> Assign "float",<:expr< try float_of_string s with [ Failure
       "float_of_string" -> raise (Parseopt.ConversionFailure (s,"float",$str:name$)) ] >>
       | `Type <:ctyp< unit  >> -> Action,<:expr< () >>
       | `Type <:ctyp< bool  >> -> Assign "boolean",<:expr< 
         try bool_of_string s with
          [ Invalid_argument "bool_of_string" -> 
            raise (Parseopt.ConversionFailure (s,"bool",$str:name$))
         ] >>
       | _ -> Assign "string",<:expr< s >>
     in
     let action = match default with 
        Some _ -> action | None ->  <:expr< Some $action$ >> in
     actiontype,
     match actiontype with 
     | Action -> 
         <:expr< Parseopt.Action (fun ap -> {(ap.options) with $lid:name$ = $action$ })>>
     | _ -> 
         <:expr< Parseopt.Conversion (fun ap s -> {(ap.options) with $lid:name$ = $action$})>>)
  in
  let arg = match actiontype with Set | Clear | Store _ | Action | Handler -> 
    <:expr< None >>
  | _ -> <:expr< Some $str:String.uppercase namestr$ >> in
  let help = try List.assoc "help" pvl with Not_found ->
    let help = 
      match actiontype with
      | Set -> "Set " ^ namestr
      | Clear -> "Clear " ^ namestr
      | Store v -> "Store " ^ v ^ " in " ^ namestr
      | Assign t -> "Assign " ^ t ^ " " ^ (String.uppercase namestr) 
        ^ " to " ^ namestr
      | Action -> "Take action " ^ namestr
      | Conversion -> "Convert " ^ (String.uppercase namestr) ^ 
        " and store in " ^ namestr
      | Handler -> "" (* no way to know what this does *) in
    let help = match default with None -> help
    | Some e -> Printf.sprintf "%s (default %s)" help (string_of_expr type_ e) 
      in
    <:expr< $str:help$ >> in
  <:expr< ( $str:namestr$, $key$, $help$, $arg$, $action$ ) >>

(** generate record type based on the param-value-annotated type.
    We need to strip param-value annotations,
    and also convert types to 'option' if no default value is provided *)
let gen_record loc (l,n,m,t,pvl,default) =
  let t = match t with 
      `Type t -> t
    | `Symbol rfl -> 
        let rfl = List.map (fun (loc,id,_) -> MLast.RfTag(id,true,[])) rfl in
        <:ctyp< [ = $list:rfl$ ] >> in
  l,n,m,(match default with None -> <:ctyp< option $t$ >> | _ -> t)

(** generate the default value of a field *)
let gen_default loc (l,n,m,t,pvl,default) =
  <:patt< $lid:n$ >>, match default with None -> <:expr< None >> | Some d -> d

EXTEND
  GLOBAL: Pcaml.expr;
  unary_constructor_declaration:
    [ [ "`"; ci = UIDENT -> (loc, ci, None) 
       | "`"; ci = UIDENT; s = STRING -> (loc, ci, Some s)
      ] ];
  pv_item:
    [ [i = LIDENT; "="; e = Pcaml.expr LEVEL "expr1"; ";" -> (i,e) ] ];
  pv_label_declaration:
    [ [ mf = OPT "mutable"; i = LIDENT; 
        default = OPT [ "="; default = Pcaml.expr LEVEL "expr1" -> default];  
        ":";  l = LIST0 pv_item;  t = 
          [ "["; OPT "|"; rfl = LIST1 unary_constructor_declaration
            SEP "|"; "]" -> 
              let rfl = List.map (fun (l,n,s) ->
                l,n,
                match s with None -> String.lowercase n | Some s -> s) rfl in
              `Symbol rfl;
           |t = Pcaml.ctyp LEVEL "simple" -> `Type t
          ]
    -> (loc, Some i, Pa_o.o2b mf, t,l,default)
      | "_"; ":";  l = LIST0 pv_item;  t = Pcaml.ctyp LEVEL "simple"
    -> (loc, None, false, `Type t,l,None)
    ]];
  pv_label_declarations:
    [ [ ld = pv_label_declaration; ";"; ldl = SELF -> (ld :: ldl)
      | ld = pv_label_declaration; ";" -> [ld]
      | ld = pv_label_declaration -> [ld] ] ];
  Pcaml.str_item: [
    (* declare an option module *)
    [ "type"; LIDENT "option"; 
       id=LIDENT; "="; "{"; fields=pv_label_declarations; "}" ->
        let named_fields = (* remove anonymous fields *)
          List.fold_left (fun acc -> function (_,None,_,_,_,_) -> acc 
          | (l,Some n,m,t,p,d) -> (l,n,m,t,p,d)::acc) [] fields in
        <:str_item<
        declare
        type $lid:id$ = {$list:List.map (gen_record loc) named_fields$};
        value $lid:id$ = {
          (Parseopt.option_defaults) with
          Parseopt.default = { $list:List.map (gen_default loc) named_fields$ };
          Parseopt.specs = $Pa_o.mklistexp loc None (List.map gen_spec fields)$;
          Parseopt.to_string = $gen_to_string loc named_fields$
        }; end >>
  ]];
END;;

(** $Revision *)
